#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <stdarg.h>

#include <pthread.h>
#include <signal.h>
#include <setjmp.h>

template <typename T>
class TlsValue
{
public:
	TlsValue()
	{
		int rc = pthread_key_create(&key, NULL);
		if (rc)
			throw ("pthread_key_create");
	}

	const T get()
	{
		// We use double C-style cast to allow using scalar datatypes
		// with sizes up to size of pointer without warnings
		return (T) pthread_getspecific(key);
	}
	void set(const T value)
	{
		int rc = pthread_setspecific(key, (void*) value);
		if (rc)
			throw ("pthread_setspecific");
	}
	~TlsValue()
	{
		int rc = pthread_key_delete(key);
		if (rc)
			throw ("pthread_key_delete");
	}
private:
	pthread_key_t key;
};

# define TLS_DECLARE(TYPE, NAME) TlsValue<TYPE> NAME
# define TLS_GET(NAME) (NAME).get()
# define TLS_SET(NAME, VALUE) (NAME).set(VALUE)

# define TLS_CLASS TlsValue


static TLS_DECLARE(sigjmp_buf*, sigjmp_ptr);
static void longjmp_sig_handler(int);
// Here we can't use atomic counter instead mutex/counter pair - or some thread may leave sync_signals_set()
// before signals are actually set in the other thread, which incremented counter first
static int sync_enter_counter = 0;

void fb_sigset(int signum, void (*handler)(int))
{
	struct sigaction act;
	memset(&act, 0, sizeof act);
	act.sa_handler = handler;
	if (sigaction(signum, &act, NULL) != 0)
		throw "sigaction";
}


void sync_signals_set(sigjmp_buf* sigenv)
{
/**************************************
 *
 *	T H D _ s y n c _ s i g n a l s _ s e t ( U N I X )
 *
 **************************************
 *
 * Functional description
 *	Set all the synchronous signals for a particular thread
 *
 **************************************/
	TLS_SET(sigjmp_ptr, sigenv);

	if (sync_enter_counter++ == 0)
	{
		printf("Set signals\n");

		fb_sigset(SIGILL, longjmp_sig_handler);
		fb_sigset(SIGFPE, longjmp_sig_handler);
		fb_sigset(SIGBUS, longjmp_sig_handler);
		fb_sigset(SIGSEGV, longjmp_sig_handler);
	}
}


void sync_signals_reset()
{
/**************************************
 *
 *	s y n c _ s i g n a l s _ r e s e t ( U N I X )
 *
 **************************************
 *
 * Functional description
 *	Reset all the synchronous signals for a particular thread
 * to default.
 *
 **************************************/

	if (--sync_enter_counter == 0)
	{
		fb_sigset(SIGILL, SIG_DFL);
		fb_sigset(SIGFPE, SIG_DFL);
		fb_sigset(SIGBUS, SIG_DFL);
		fb_sigset(SIGSEGV, SIG_DFL);
	}
}

static void longjmp_sig_handler(int sig_num)
{
/**************************************
 *
 *	l o n g j m p _ s i g _ h a n d l e r
 *
 **************************************
 *
 * Functional description
 *	The generic signal handler for all signals in a thread.
 *
 **************************************/

	siglongjmp(*TLS_GET(sigjmp_ptr), sig_num);
}

int main(int ac, char** av)
{
	pthread_t thread = 0xdeadbeef;

	try
	{
			{
				sigjmp_buf sigenv;
				if (sigsetjmp(sigenv, 1) == 0)
				{
					sync_signals_set(&sigenv);
/*					int* iptr = nullptr;
					if (ac == 1)
						*iptr = 0;
*/
					if (pthread_kill(thread, 0) == ESRCH)
					{
						// Thread does not exist any more
						printf("pthread_kill works\n");
					}
				}
				else
				{
					// segfault in pthread_kill() - thread does not exist any more
					printf("Sigenv works\n");
				}
				sync_signals_reset();
			}
	}
	catch(const char* text)
	{
		fprintf(stderr, "%s\n", text);
		return 1;
	}

	return 0;
}

